library ieee;
use ieee.std_logic_1164.all;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

library UNISIM;
use UNISIM.VComponents.all;

entity opb_ak4565 is
  
  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    C_BASEADDR   : std_logic_vector(0 to 31) := X"00000000";
    C_HIGHADDR   : std_logic_vector(0 to 31) := X"FFFFFFFF");

  port (
    OPB_Clk     : in  std_logic;
    OPB_Rst     : in  std_logic;
    OPB_ABus    : in  std_logic_vector(0 to C_OPB_AWIDTH-1);
    OPB_BE      : in  std_logic_vector(0 to C_OPB_DWIDTH/8-1);
    OPB_DBus    : in  std_logic_vector(0 to C_OPB_DWIDTH-1);
    OPB_RNW     : in  std_logic;
    OPB_select  : in  std_logic;
    OPB_seqAddr : in  std_logic;        -- Sequential Address
    Sln_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sln_errAck  : out std_logic;        -- (unused)
    Sln_retry   : out std_logic;        -- (unused)
    Sln_toutSup : out std_logic;        -- Timeout suppress
    Sln_xferAck : out std_logic;

    AU_CSN      : out std_logic;
    AU_BCLK     : out std_logic;
    AU_MCLK     : out std_logic;
    AU_LRCK     : out std_logic;
    AU_SDTI     : out std_logic;
    AU_SDTO     : in  std_logic);
  
end opb_ak4565;

architecture behavioral of opb_ak4565 is
  component fifo_256x16
    port (
      din: IN std_logic_VECTOR(15 downto 0);
      wr_en: IN std_logic;
      wr_clk: IN std_logic;
      rd_en: IN std_logic;
      rd_clk: IN std_logic;
      ainit: IN std_logic;
      dout: OUT std_logic_VECTOR(15 downto 0);
      full: OUT std_logic;
      empty: OUT std_logic);
  end component;

signal count3 : std_logic_vector(11 downto 0);
signal count : std_logic_vector(4 downto 0);
signal shiftin, shiftout, dshiftin : std_logic_vector(15 downto 0) :=X"0000";
signal rec, play, audio_bram_clk,audio_bram_adcdata : std_logic;
signal adcdone, dacload, wr : std_logic;
signal dacout, adcout, adc_mux : std_logic_vector(15 downto 0);
signal addr, wdata, rdata : std_logic_vector(31 downto 0);
signal wr_bram : std_logic;
signal addr_bramadc, addr_bramdac : std_logic_vector(7 downto 0);
signal dac_full, adc_empty : std_logic;
signal adc_full, dac_empty : std_logic;
signal cs, rnw, rd_bram : std_logic;

signal bclk,mclk,lrclk : std_logic;


type opb_state is (IDLE, COMMON, XREAD, XFER);
signal cstate, nxstate : opb_state;

signal ld_sldbus : std_logic;

begin  -- behavioral

-- opb stuff

process(OPB_Rst, OPB_Clk)
begin
  if OPB_CLk'event and OPB_clk='1' then
    cstate <= nxstate;
    wdata <= OPB_DBus;
    addr <= OPB_ABus;
    if ld_sldbus = '1' then
      Sln_DBus(0 to 15) <= X"0000";
      Sln_DBus(16 to 31) <= adc_mux;
    else
      Sln_DBus <= (others => '0');
    end if;
    rnw <= OPB_RNW;
  end if;
  if OPB_Rst = '1' then
    cstate <= IDLE;
    wdata <= (others => '0');
    Sln_DBus <= (others => '0');
  end if;
end process;  
 adc_mux <= adcout when addr(2 downto 0) = 0
           else X"000" & dac_full & dac_empty & adc_full & adc_empty;
cs <= '1' when OPB_ABus(0 to 15) = C_BASEADDR(0 to 15) and OPB_Select = '1' else '0';
-- DAC FIFO  
dac_fifo : fifo_256x16
  port map (
    din => wdata(15 downto 0),
    wr_en => wr_bram,
    wr_clk => OPB_Clk,
    rd_en => dacload,
    rd_clk => count(4),
    ainit => OPB_Rst,
    dout => dacout,
    full => dac_full,
    empty => dac_empty
    );

--ADC FIFO
adc_fifo : fifo_256x16
  port map (
    din => rdata(15 downto 0),
    wr_en => wr,
    wr_clk => bclk,
    rd_en => rd_bram,
    rd_clk => OPB_Clk,
    ainit => OPB_Rst,
    dout => adcout,
    full => adc_full,
    empty => adc_empty
    );

process(cstate, cs, rnw)
begin
  nxstate <= cstate;
  wr_bram <= '0';
  Sln_xferAck <= '0';
  ld_sldbus <= '0';
  rd_bram <= '0';
  
  case cstate is
    when IDLE =>
      if cs = '1' then
        nxstate <= COMMON;
      end if;
    when COMMON =>
      if rnw = '0'  then
        wr_bram <= '1';
        nxstate <= XFER;
      else
        nxstate <= XREAD;
        if addr(2 downto 0) = 0 then
          rd_bram <= '1';
        end if;
      end if;

    when XREAD => ld_sldbus <= '1';
                  nxstate <= XFER;
      
    when xfer => Sln_xferAck <= '1';
                 NXSTATE <= IDLE;
                 
  end case;

  
end process;  


  
  
AU_MCLK <= mclk;
AU_BCLK <= bclk;
AU_LRCK <= lrclk;
AU_CSN <= '1';

rec <= OPB_select and OPB_RNW;
--wr_bram <= OPB_select and not OPB_RNW;
addr_bramadc <= count3(11 downto 4) - 1;
addr_bramdac <= count3(11 downto 4) + 1;

    Sln_errAck <= '0';
    Sln_retry <= '0';
    Sln_toutSup <= '0';
--    Sln_xferAck <= '0';

    
-- CLOCK Generators  -------------------------------------------------------

process(OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      count <= "00000";
    elsif OPB_Clk'event and OPB_Clk = '1'  then
      count <= count + 1;
    end if;
end process;
mclk <= count(1);
bclk <= count(4);

  process(OPB_Rst, bclk)
  begin
    if OPB_Rst = '1' then
      count3 <= X"000";
      lrclk <= '0';
    elsif bclk'event and bclk = '0' then
      count3 <= count3 + 1;

      if count3(4 downto 0) = 31 then
        lrclk <= '1';
      elsif count3(4 downto 0) = 15 then
        lrclk <= '0';
      end if;

      if count3(3 downto 0) = 15  then
        wr  <= '1';
      else
        wr <= '0';
      end if;
      
    end if;
  end process; 

  
  

  
process(OPB_Rst, count(4))
  begin
    if OPB_Rst = '1'  then
      shiftin <= X"0000";
      rdata <= X"00000000";
    elsif count(4)'event and count(4) = '0' then
      if count3(3 downto 0) = 0 then
        rdata(15 downto 0) <= shiftin;
        rdata(31 downto 16) <= X"0000";
      end if;
        shiftin(15) <= shiftin(14);
        shiftin(14) <= shiftin(13);
        shiftin(13) <= shiftin(12);
        shiftin(12) <= shiftin(11);
        shiftin(11) <= shiftin(10);
        shiftin(10) <= shiftin(9);
        shiftin(9) <= shiftin(8);
        shiftin(8) <= shiftin(7);
        shiftin(7) <= shiftin(6);
        shiftin(6) <= shiftin(5);
        shiftin(5) <= shiftin(4);
        shiftin(4) <= shiftin(3);
        shiftin(3) <= shiftin(2);
        shiftin(2) <= shiftin(1);
        shiftin(1) <= shiftin(0);
        shiftin(0) <= AU_SDTO;
    end if;
  end process;    

  process(OPB_Rst, count(4))
  begin
    if OPB_Rst = '1' then
      shiftout <= X"0000";
      AU_SDTI <= '0';
    elsif count(4)'event and count(4) = '0' then
      dacload <= '0';
         
      if count3(3 downto 0) = 14 then
        dacload <= '1';
      end if;
         
      if count3(3 downto 0) = 14 then
        shiftout <= dacout;
      else
        shiftout <=  shiftout(14 downto 0)  & '0';
      end if;

      AU_SDTI <= shiftout(15);

    
    end if;
       
  end process;
        
  
end behavioral;

